from multigraph import *
from calc_strat import Strat
from IOSM import *
import random
class Oracle:
    def __init__(self, MAX_NODE_VISITED: int, x: list[int],g: BipartiteMultiGraph):
        self.g = g
        self.MAX_NODE_VISITED = MAX_NODE_VISITED
        self.current_node_visited = 0
        self.phase = "pre"
        self.u = g.idle
        self.x_idx = 0
        self.transis = g.transis

        if self.u == x[0]:
            self.phase = "during"
        self.x = x
        self.n_x = len(x)
        self.current_node_visited = 0
        self.strat = Strat(x[0], self.g)
        self.strat.calc_w()
        self.strat.calc_rank()

    def update_phases(self):
        if(self.phase == "pre"):
            if(len(self.x) == 1 and self.u == self.x[0]): #on entre dans la phase during et direct apres c la phase finale.
                self.phase = "post"
                self.strat = Strat(self.g.idle, self.g)
                self.strat.calc_w()
                self.strat.calc_rank()
            elif(self.u == self.x[0]):
                self.phase = "during"
                self.x_idx = 0

        elif(self.phase == "during"):
            if(self.x[self.x_idx] == self.u): ## si on reste dans la phase intermediaire

                if(self.x_idx == self.n_x -1):  #on a fini la sequence
                    self.phase = "post"
                    self.strat = Strat(self.g.idle, self.g)
                    self.strat.calc_w()
                    self.strat.calc_rank()
                    

            else:
                self.phase = "pre"
            #ne pas retirer la fonction dans update move car elle verifie si on reste bien dans la sequence des x. ici pas necessaire
        else:#a la fin, on s'en fiche...
            pass


    def update_move(self, transi:str) -> tuple[str, State, list[int|bool]] :#quand le joueur adverse joue un coup
        if(transi!="__NotMyTurn"):
            self.current_node_visited+=1
            for v in self.g.E[self.u]:
                if transi in self.g.E_label[self.u,v]:
                    self.u = v
                    break
            
            if(self.phase == "during"):
                self.x_idx+=1
            self.update_phases()
        else:
            return self.next_move()

    
    def check_pass(self) -> bool:
        return self.phase == "post" and self.u == self.g.idle

    def check_fail(self) -> bool:
        return self.u == "bot"
    
    def check_win(self) -> bool:
        return self.check_pass() or self.check_fail()
        
    def next_move(self) -> tuple[str, State, list[int|bool]]:
        self.current_node_visited+=1
        transi_taken = ""
        v = 0
        if(self.phase == "pre"):
            v =  self.strat.strategy(self.u)
        elif(self.phase == "during"):
            self.x_idx+=1
            v = self.x[self.x_idx]
        elif(self.phase == "post"):
            v = self.strat.strategy(self.u)
        transi_taken = self.g.E_label[self.u,v][0]
        output_var = self.output_variables(self.u, v, transi_taken)

        if(self.g.E_label[self.u,v] == "tau" and self.strat.rank[v]!=(float("inf"), float("inf"))): #on a est resté dans un état interne...
            self.u = v
            self.update_phases()
            return None
        
        self.u = v
        self.update_phases()
        return transi_taken, self.u, output_var

    def output_variables(self, u, v, t_name):
        context = {}
        max_attempts = 10000
        if(u, v, t_name) in self.transis:
            t = self.transis[u, v, t_name]

            for _ in range(max_attempts):
                for var in t.domains :
                    value = generate_value(var.var_type)
                    context[var.var_name] = value
                if evaluate_conditions(t, context):
                    return context

            raise ValueError(f"Aucune liste de guards valide trouvée après {max_attempts} tentatives")
        
        raise ValueError(f"Erreur : état non accessible")    
    
def generate_value(t):
    if t == int:
        return random.randint(0, 10000)
    elif t == bool:
        return random.choice([True, False])
    elif t == float:
        return round(random.uniform(0, 10000.0), 2)
    else:
        raise TypeError(f"Type non supporté : {t}")
    