from dataclasses import dataclass
from IOSM import State, IOSM_from_XLIA, IOSM, TransitionType, Transition
import networkx as nx
import sys
import matplotlib.pyplot as plt
import copy
label = str

@dataclass
class Graph:
    """Graph that represents an accessibility game"""
    Vb: list[State]
    Vr: list[State]
    E : dict[State, list[State]]
    E_label : dict[tuple[State, State], Transition]
    idle: State
    transitions : set[Transition]
    
    
    def render_graph(self)->None:
        """Render the graph using networkx and matplotlib"""
        G = nx.DiGraph()
        G.add_nodes_from(self.Vb, bipartite=0)
        G.add_nodes_from(self.Vr, bipartite=1)
        for s, targets in self.E.items():
            for target in targets:
                G.add_edge(s, target, label=self.E_label[(s, target)])
        
        color_map = []
        for node in G.nodes():
            if node == self.idle:
                color_map.append("green")
            else:
                color_map.append("blue" if node in self.Vb else "red")
        
        pos = nx.spring_layout(G,k=30, seed=42)
        
        nx.draw_networkx_nodes(G, pos, node_color=color_map)
        nx.draw_networkx_labels(G, pos)
        nx.draw_networkx_edges(
            G,
            pos,
            edge_color="gray",
            arrows=True,
        )
        edge_labels = {(s, t): data['label'] for s, t, data in G.edges(data=True)}
        nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels, label_pos=0.8)
        if not nx.algorithms.is_bipartite(G):
            print("Warning: graph is not bipartite")
        plt.show()
    
    
      
class BipGraph(Graph):

    def __init__(self, iosm: IOSM)-> None:

        self.cnt = 0
        a, b, c, d, e,f = self.transform(iosm)
        self.n = len(a) + len(b)
        super().__init__(a, b, c, d, e, f)

    def transform(self, iosm: IOSM):
        
        
        
        print("transitions ", iosm.transitions)
        nodes = copy.deepcopy(iosm.states)
        #nodes.add(iosm.idle_state)
        E = {state: [] for state in iosm.states}
    # E[ios.]=[]
        E_label = {}
        for transit in iosm.transitions :
            print(transit.from_node, transit.to_node)
            if(not transit.to_node in E[transit.from_node]):
                E[transit.from_node].append(transit.to_node)
            E_label[transit.from_node, transit.to_node] = ("?" if transit.t_type==TransitionType.INPUT else "!") + transit.name
        
        print("E_label", E_label)
        print("E", E)
        nodes = list(nodes)
        print("STATES", nodes)

        node_counter = len(nodes)
        found_bad : bool = True

        while found_bad:
            found_bad = False
            for u in nodes:
                for v in E[u]:
                    for w in E[v]:
                        print(E_label)
                        print(E_label[u,v], E_label[v, w])
                        if(E_label[u, v][0] == E_label[v,w][0] and E_label[u, v]!="tau"):

                            E[node_counter] = []
                            E[node_counter].append(v)

                            E_label[u, node_counter] = E_label[u, v]
                            E_label[node_counter, v] = "tau"
    
                            E[u].remove(v)
                            E[u].append(node_counter)

                            del E_label[u,v]
                            found_bad = True
                            node_counter+=1

                            break
                    if found_bad:
                        break
                if found_bad: 
                    break


        print(E_label)

        V = [set(), set()]

        self.cnt = node_counter


        G = nx.DiGraph()
        G.add_nodes_from(nodes, bipartite=0)
        for s, targets in E.items():
            for target in targets:
                t = E_label[(s, target)]
                
                G.add_edge(s, target, label=t,)
            
    
        while(not nx.algorithms.is_bipartite(G)):
            self.color(E, E_label, V, iosm.idle_state,-1, 0)
            G = nx.DiGraph()
            G.add_nodes_from(nodes, bipartite=0)
            for s, targets in E.items():
                for target in targets:
                    t = E_label[(s, target)]
                    
                    G.add_edge(s, target, label=t,)
                
        
            V[0] = set()
            V[1] = set()


        self.color(E, E_label, V, iosm.idle_state, -1, 0)
        V[1].add("err")
        V[0].add("bot")
        E["err"]=[]
        E["bot"]=[]
        for r in V[1]:
            E[r].append("bot")
            E_label[r,"bot"]="tau_bot"
        E["err"].append("bot")
        E["bot"].append("err")
        E_label["err", "bot"] = "tau_bot"
        E_label["bot","err" ]="tau"
        nodes.append("bot")
        nodes.append("err")
        #g = Graph(list(V[0]), list(V[1]), E, E_label, iosm.idle_state)
       # return E, E_label, nodes, iosm.idle_state


        g= Graph(list(V[0]), list(V[1]), E, E_label, iosm.idle_state, iosm.transitions)
        return g.Vb, g.Vr, E, E_label, iosm.idle_state, iosm.transitions

    def color(self, E, E_label,V,  u, p,  i):
        
        if(u in V[i]):
            return
        elif(u in V[(i+1)%2]):
            E[p].append(self.cnt)
            E[self.cnt] = []
            E[self.cnt].append(u)
            V[i].add(self.cnt)
            E[p].remove(u)
            E_label[p, self.cnt]= E_label[p, u]
            E_label[self.cnt, u]="tau"
            del E_label[p, u]
            print("BAD NODE :", u)
            self.cnt+=1
            return
        V[i].add(u)
        for v in E[u]:
            self.color(E, E_label, V, v,u, (i+1)%2)

        
