import json
from dataclasses import dataclass
from enum import Enum
import re

State = str
Interaction = str
@dataclass
class TransitionType(Enum):
    INPUT = "input"
    OUTPUT = "output"
    INTERNAL = "internal" #tau

@dataclass
class Transition:
    name: str
    t_type: TransitionType
    guard_condition: str
    from_node: State
    to_node: State
    
    def __hash__(self):
        return hash((self.name, self.t_type, self.guard_condition, self.from_node, self.to_node))

@dataclass
class IOSM:
    states: set[State]
    transitions: set[Transition]
    idle_state: State



def json_to_IOSM(filepath) -> IOSM:
    with open(filepath, "r") as json_data:
        data = json.load(json_data)
    data_states=data["TESTCASE"]["states"]
    inputs=[i["name"] for i in data["SUT"]["input_ports"]]
    outputs=[i["name"] for i in data["SUT"]["output_ports"]]
    #s_p = []
    #s_m = []
    states= set()
    transitions = set()

    for s in data_states:
        states.add(s["name"])
        for trans in s["transitions"]:
            if trans["action"] in inputs:
                t_type=TransitionType.INPUT
            elif trans["action"] in outputs:
                t_type=TransitionType.OUTPUT
            else:
                t_type=TransitionType.INTERNAL
            if "next" in trans.keys():
                t_to=trans["next"]
            else:
                t_to = "bot"

            transitions.add(
                Transition(
                    name=trans["name"],
                    t_type=t_type,
                    from_node=s["name"],
                    to_node=t_to
                ))
            
    
    #The idle state is the first state in the json file
    idle_state = list(data_states.keys())[0]

    return IOSM(states=states, transitions=transitions, idle_state=idle_state)

    #print(states)

if __name__ == "__main__":
    x = json_to_IOSM("testcases.json")
    print(x)